from tests.declarativeunittest import *
from construct import *
from construct.lib import *

from gallery import pe32file, UTIndex


def test_pe32():
    commondump(pe32file, "python37-win32.exe")
    commondump(pe32file, "python37-win64.exe")
    commondump(pe32file, "SharpZipLib0860-dotnet20.dll")
    commondump(pe32file, "sqlite3.dll")

def test_utindex():
    d = UTIndex()
    test_data = [
        [0x0f],  # 0x0f
        [0x4f, 0x40],  # (0x40 << 6) + 0x0f = 0x100f
        [0x8f],  # -0x0f
        [0xcf, 0x40],  # -((0x40 << 6) + 0x0f) = -0x100f
        [0x4f, 0x80, 0x40],  # (0x40 << 13) + 0x0f = 0x8000f
        [0x4f, 0x80, 0x80, 0x40],  # (0x40 << 20) + 0x0f = 0x400000f
        [0x4f, 0x80, 0x80, 0x80, 0x8f],  # 0x8f << 27 + 0x0f = 0x47800000f
    ]
    expected_values = [
        0x0f,
        0x100f,
        -0x0f,
        -0x100f,
        0x8000f,
        0x400000f,
        0x47800000f,
    ]
    for test, ev in zip(test_data, expected_values):
        assert d.parse(bytes(test)) == ev
        assert d.build(ev) == bytes(test)
    assert raises(d.sizeof) == SizeofError



from construct import *
from construct.lib import *


class UTIndex(Construct):
    """
    Format for "Index" objects in Unreal Tournament 1999 packages.
    Index objects are variable length signed integers with the following structure:

    +------------------------------------+-------------------------+--------------+
    | Byte 0                             | Bytes 1-3               | Byte 4       |
    +----------+----------+--------------+----------+--------------+--------------+
    | Sign Bit | More Bit | Data Bits[6] | More Bit | Data Bits[7] | Data Bits[8] |
    +----------+----------+--------------+----------+--------------+--------------+

    If the "More" bit is 0 in any byte, that's the end of the Index. Otherwise,
    keep going. There cannot be more than 5 bytes in an Index so Byte 4 doesn't
    have a "More" bit.
    """
    lengths = {0: 6, 1: 7, 2: 7, 3: 7, 4: 8}
    negative_bit = 0x80

    @staticmethod
    def _get_data_mask(length):
        return (0xFF ^ (0xFF << length)) & 0xFF

    @staticmethod
    def _get_more_bit(length):
        return 1 << length

    def _parse(self, stream, context, path):
        result = 0
        sign = 1
        i = 0
        depth = 0
        while True:
            length = self.lengths[i]
            bits = byte2int(stream_read(stream, 1, path))
            mask = self._get_data_mask(length)
            data = bits & mask
            more = self._get_more_bit(length) & bits
            if (i == 0) and (self.negative_bit & bits):
                sign = -1
            result |= data << depth
            if not more:
                break
            i += 1
            depth += length
        return sign * result

    def _build(self, obj, stream, context, path):
        if not isinstance(obj, integertypes):
            raise IntegerError("Value is not an integer")
        to_write = obj
        for i in range(5):
            byte = 0
            length = self.lengths[i]
            if i == 0:
                negative = obj < 0
                byte |= self.negative_bit * negative
                if negative:
                    to_write *= -1
            mask = self._get_data_mask(length)
            byte |= to_write & mask
            to_write >>= length
            more_bit = (to_write > 0) and self._get_more_bit(length)
            byte |= more_bit
            byte &= 0xFF
            stream_write(stream, int2byte(byte), 1, path)
            if not more_bit:
                break
        return obj



import pytest

xfail = pytest.mark.xfail
skip = pytest.mark.skip
skipif = pytest.mark.skipif

import os, math, random, collections, itertools, io, hashlib, binascii

from construct import *
from construct.lib import *

if not ONWINDOWS:
    devzero = open("/dev/zero", "rb")

ident = lambda x: x

def raises(func, *args, **kw):
    try:
        return func(*args, **kw)
    except Exception as e:
        return e.__class__

def common(format, datasample, objsample, sizesample=SizeofError, **kw):
    # following are implied (re-parse and re-build)
    # assert format.parse(format.build(obj)) == obj
    # assert format.build(format.parse(data)) == data
    obj = format.parse(datasample, **kw)
    assert obj == objsample
    data = format.build(objsample, **kw)
    assert data == datasample

    if isinstance(sizesample, int):
        size = format.sizeof(**kw)
        assert size == sizesample
    else:
        size = raises(format.sizeof, **kw)
        assert size == sizesample

    # attemps to compile, ignores if compilation fails
    # following was added to test compiling functionality
    # and implies: format.parse(data) == cformat.parse(data)
    # and implies: format.build(obj) == cformat.build(obj)
    try:
        cformat = format.compile()
    except Exception:
        pass
    else:
        obj = cformat.parse(datasample, **kw)
        assert obj == objsample
        data = cformat.build(objsample, **kw)
        assert data == datasample

def commonhex(format, hexdata):
    commonbytes(format, binascii.unhexlify(hexdata))

def commondumpdeprecated(format, filename):
    filename = "tests/deprecated_gallery/blobs/" + filename
    with open(filename,'rb') as f:
        data = f.read()
    commonbytes(format, data)

def commondump(format, filename):
    filename = "tests/gallery/blobs/" + filename
    with open(filename,'rb') as f:
        data = f.read()
    commonbytes(format, data)

def commonbytes(format, data):
    obj = format.parse(data)
    data2 = format.build(obj)



from tests.declarativeunittest import *
from construct.lib.binary import *


def test_integer2bits():
    assert integer2bits(0, 0, False) == b""
    assert integer2bits(0, 0, True) == b""
    assert integer2bits(19, 5) == b"\x01\x00\x00\x01\x01"
    assert integer2bits(19, 8) == b"\x00\x00\x00\x01\x00\x00\x01\x01"
    assert integer2bits(-13, 5, True) == b"\x01\x00\x00\x01\x01"
    assert integer2bits(-13, 8, True) == b"\x01\x01\x01\x01\x00\x00\x01\x01"
    assert raises(integer2bits, 0, -1) == ValueError
    assert raises(integer2bits, -1, 8, False) == ValueError
    assert raises(integer2bits, -2**64, 8, True) == ValueError
    assert raises(integer2bits,  2**64, 8, True) == ValueError
    assert raises(integer2bits, -2**64, 8, False) == ValueError
    assert raises(integer2bits,  2**64, 8, False) == ValueError

def test_integer2bytes():
    assert integer2bytes(0, 0, False) == b""
    assert integer2bytes(0, 0, True) == b""
    assert integer2bytes(0, 4) == b"\x00\x00\x00\x00"
    assert integer2bytes(1, 4) == b"\x00\x00\x00\x01"
    assert integer2bytes(19, 4) == b'\x00\x00\x00\x13'
    assert integer2bytes(255, 1) == b"\xff"
    assert integer2bytes(255, 4) == b"\x00\x00\x00\xff"
    assert integer2bytes(-1, 4, True) == b"\xff\xff\xff\xff"
    assert integer2bytes(-255, 4, True) == b"\xff\xff\xff\x01"
    assert raises(integer2bytes, 0, -1) == ValueError
    assert raises(integer2bytes, -1, 8, False) == ValueError
    assert raises(integer2bytes, -2**64, 4, True) == ValueError
    assert raises(integer2bytes,  2**64, 4, True) == ValueError
    assert raises(integer2bytes, -2**64, 4, False) == ValueError
    assert raises(integer2bytes,  2**64, 4, False) == ValueError

def test_bits2integer():
    assert bits2integer(b"", False) == 0
    assert bits2integer(b"", True) == 0
    assert bits2integer(b"\x01\x00\x00\x01\x01", False) == 19
    assert bits2integer(b"\x01\x00\x00\x01\x01", True) == -13

def test_bytes2integer():
    assert bytes2integer(b"", False) == 0
    assert bytes2integer(b"", True) == 0
    assert bytes2integer(b"\x00") == 0
    assert bytes2integer(b"\x00", True) == 0
    assert bytes2integer(b"\xff") == 255
    assert bytes2integer(b"\xff", True) == -1
    assert bytes2integer(b'\x00\x00\x00\x13', False) == 19
    assert bytes2integer(b'\x00\x00\x00\x13', True) == 19

def test_cross_integers():
    for i in [-300,-255,-100,-1,0,1,100,255,300]:
        assert bits2integer(integer2bits(i,64,signed=(i<0)),signed=(i<0)) == i
        assert bytes2integer(integer2bytes(i,8,signed=(i<0)),signed=(i<0)) == i
        assert bits2bytes(integer2bits(i,64,signed=(i<0))) == integer2bytes(i,8,signed=(i<0))
        assert bytes2bits(integer2bytes(i,8,signed=(i<0))) == integer2bits(i,64,signed=(i<0))

def test_bytes2bits():
    assert bytes2bits(b"") == b""
    assert bytes2bits(b"ab") == b"\x00\x01\x01\x00\x00\x00\x00\x01\x00\x01\x01\x00\x00\x00\x01\x00"

def test_bits2bytes():
    assert bits2bytes(b"") == b""
    assert bits2bytes(b"\x00\x01\x01\x00\x00\x00\x00\x01\x00\x01\x01\x00\x00\x00\x01\x00") == b"ab"
    assert raises(bits2bytes, b"\x00") == ValueError
    assert raises(bits2bytes, b"\x00\x00\x00\x00\x00\x00\x00") == ValueError

def test_swapbytes():
    assert swapbytes(b"") == b""
    assert swapbytes(b"abcd") == b"dcba"

def test_swapbytesinbits():
    assert swapbytesinbits(b"") == b""
    assert swapbytesinbits(b"0000000011111111") == b"1111111100000000"
    assert raises(swapbytesinbits, b"1") == ValueError

def test_swapbitsinbytes():
    assert swapbitsinbytes(b"") == b""
    assert swapbitsinbytes(b"\xf0") == b"\x0f"
    assert swapbitsinbytes(b"\xf0\x00") == b"\x0f\x00"



from tests.declarativeunittest import *
from construct.lib.hex import *


def test_hexdump():
    for i in range(100):
        assert hexundump(hexdump(b"?"*i,32),32) == b"?"*i

def test_hexundump_issue_882():
    data1 = (hexdump(hexundump(
"""
0000   30 31 32 33 34 35 36 5C 0123456\\
0008   38                      8

""", linesize=8), linesize=8))

    data2 = (hexdump(hexundump(
"""
0000   30 31 32 33 34 35 36 37 01234567
0008   38                      8

""", linesize=8), linesize=8))

    print(data1)
    print(data2)
    assert data1 == hexdump(b"0123456\\8", 8)
    assert data2 == hexdump(b"012345678", 8)
